package be.iminds.ilabt.jfed.experimenter_gui.bugreporting;

import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.editor.properties.AdaptivePropertiesDialog;
import be.iminds.ilabt.jfed.experimenter_gui.ui.ResourceCell;
import be.iminds.ilabt.jfed.experimenter_gui.ui.SceneAwareController;
import be.iminds.ilabt.jfed.experimenter_gui.util.StringConverters;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Resource;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.OSDetector;
import javafx.beans.binding.Bindings;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.geometry.Insets;
import javafx.scene.Scene;
import javafx.scene.control.*;
import javafx.scene.image.ImageView;
import javafx.scene.input.Clipboard;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyCodeCombination;
import javafx.scene.layout.BorderPane;
import javafx.scene.layout.Pane;
import javafx.scene.layout.StackPane;
import javafx.scene.paint.Color;
import javafx.stage.FileChooser;
import javafx.stage.Stage;
import org.controlsfx.dialog.CommandLinksDialog;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import java.io.File;
import java.util.Comparator;
import java.util.Date;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 12/9/13
 * Time: 8:58 AM
 */
public class BugReportDialog implements SceneAwareController {
    private static final Logger LOG = LoggerFactory.getLogger(BugReportDialog.class);
    private static final String BUGREPORT_EXTENSION = ".xml";

    private final BugReportComposer bugReportComposer;
    private final JFedGuiPreferences jFedPreferences;
    private final JFedGuiConfig jFedGuiConfig;
    private final TestbedInfoSource testbedInfoSource;
    private final Clipboard systemClipboard = Clipboard.getSystemClipboard();

    private final BooleanProperty submitting = new SimpleBooleanProperty(false);

    private enum Part1Choice { QUESTION, FEATURE, BUG, OTHER }
    private enum Part2Choice { JFED, TESTBED, OTHER }

    @FXML
    private BorderPane root;
    @FXML
    private StackPane centerStack;

    @FXML
    private Label titleLabel;
    @FXML
    private Pane reportTypePane;
    @FXML
    private Pane reportDetailsPane;

    @FXML
    private TextField subjectTextField;
    @FXML
    private TextArea headerTextArea;
    @FXML
    private Label descriptionLabel;
    @FXML
    private TextArea descriptionTextArea;
    @FXML
    private TextField versionTextField;
    @FXML
    private TextField environmentTextField;
    @FXML
    private TextField credentialsTextField;
    @FXML
    private TextField emailTextField;
    @FXML
    private Button saveToFileButton;
    @FXML
    private Button submitButton;
    @FXML
    private Button cancelButton;
    @FXML
    private Label includedCallsLabel;
    @FXML
    private ProgressIndicator submitProgressIndicator;
    @FXML
    private Label progressLabel;

    @FXML
    private StackPane screenshotStackPane;
    @FXML
    private CheckBox includeScreenshotCheckBox;
    @FXML
    private ImageView screenshotImageView;
    @FXML
    private Button pasteScreenshotButton;

    private BugReportToFileTask reportTask = null;


    @FXML BugReportQuestionButton part1Question;
    @FXML BugReportQuestionButton part1Feature;
    @FXML BugReportQuestionButton part1Bug;
    @FXML BugReportQuestionButton part1Other;

    @FXML BugReportQuestionButton part2jFed;
    @FXML BugReportQuestionButton part2Testbed;
    @FXML BugReportQuestionButton part2Other;

    @FXML Label part3label;
    @FXML ComboBox<Server> part3Combobox;
    @FXML Button part3ok;

    @FXML CheckBox part4AllowPublic;
    @FXML Label part4note;
    @FXML Button part4ok;

    /** part1 = 1, ...,  part4 = 4, reportDetails = 5*/
    private int currentPart = 1;

    private String fullChoiceText = "";
    private String subjectChoiceText = "";
    private Part1Choice part1Choice;
    private Part2Choice part2Choice;
    private Server part3Choice;
    private Boolean part4Choice;


    @Inject
    BugReportDialog(BugReportComposer bugReportComposer, JFedGuiPreferences jFedPreferences,
                    JFedGuiConfig jFedGuiConfig, TestbedInfoSource testbedInfoSource) {
        this.bugReportComposer = bugReportComposer;
        this.jFedPreferences = jFedPreferences;
        this.jFedGuiConfig = jFedGuiConfig;
        this.testbedInfoSource = testbedInfoSource;
    }

    @FXML
    private void initialize() {
        submitButton.disableProperty().bind(submitting);
        saveToFileButton.disableProperty().bind(submitting);
        submitProgressIndicator.visibleProperty().bind(submitting);


        descriptionTextArea.textProperty().bindBidirectional(bugReportComposer.descriptionProperty());
        emailTextField.textProperty().bindBidirectional(bugReportComposer.emailProperty());
        versionTextField.textProperty().bind(bugReportComposer.versionProperty());
        environmentTextField.textProperty().bind(bugReportComposer.environmentProperty());
        credentialsTextField.textProperty().bind(bugReportComposer.credentialsProperty());
        includedCallsLabel.textProperty().bind(Bindings.format("%d calls", bugReportComposer.callsCountProperty()));

        screenshotImageView.imageProperty().bindBidirectional(bugReportComposer.screenshotProperty());
        screenshotImageView.fitWidthProperty().bind(screenshotStackPane.widthProperty());
        screenshotImageView.fitHeightProperty().bind(screenshotStackPane.heightProperty());

        //bind screenshot-imageview to clipboard
        screenshotImageView.imageProperty().addListener(observable -> {
            if (screenshotImageView.getImage() != null) {
                includeScreenshotCheckBox.setSelected(true);
                includeScreenshotCheckBox.setDisable(false);
                pasteScreenshotButton.setVisible(false);
            }
        });

        includeScreenshotCheckBox.setOnAction(observable -> {
            screenshotImageView.setImage(null);
            includeScreenshotCheckBox.setDisable(true);
            pasteScreenshotButton.setVisible(true);
        });

        reportTypePane.getStylesheets().add(CommandLinksDialog.class.getResource("dialogs.css").toExternalForm());

        part1Question.setVisible(true);
        part2jFed.setVisible(false);
        part3label.setVisible(false);
        part4AllowPublic.setVisible(false);

        centerStack.requestLayout();


        testbedInfoSource.getServers().stream()
                .sorted(Comparator.comparing(Server::getName).thenComparing(Server::getId))
                .forEach(part3Combobox.getItems()::add);
        part3Combobox.setConverter(StringConverters.SERVER_STRING_CONVERTER);
        part3Combobox.setCellFactory(param -> new ResourceCell(null, jFedGuiConfig));
        part3Combobox.setButtonCell(new ResourceCell(null, jFedGuiConfig) {
            {
                nameLabel.setTextFill(Color.BLACK);
                container.setPadding(new Insets(0, 0, 0, 0));
            }

            @Override
            protected void updateItem(Server item, boolean empty) {
                super.updateItem(item, empty);
                LOG.debug("part3Combobox ButtonCell updateItem("+(item == null ? "null" : item.getId()+" - "+item.getName())+", "+empty+")");
            }
        });
    }

    /*
    *
    private Part1Choice part1Choice;
    private Part2Choice part2Choice;
    private Server part3Choice;
    private Boolean part4Choice;
    */

    private void showPart() {
        if (currentPart <= 0 || currentPart > 5)
            throw new RuntimeException("Invalid current part "+currentPart);

        part1Question.setVisible(currentPart == 1);
        part2jFed.setVisible(currentPart == 2);
        part3label.setVisible(currentPart == 3);
        part4AllowPublic.setVisible(currentPart == 4);

        if (currentPart >= 1 && currentPart <= 4) {
            reportTypePane.setVisible(true);
            reportDetailsPane.setVisible(false);
            submitButton.setVisible(false);
        }
        if (currentPart == 5) {
            reportTypePane.setVisible(false);
            reportDetailsPane.setVisible(true);
            submitButton.setVisible(true);

            centerStack.requestLayout();
        }
    }

    public void onPartOne(ActionEvent e) {
        if (e.getSource() == part1Question) {
            part1Choice = Part1Choice.QUESTION;
            part4AllowPublic.setSelected(true);
            fullChoiceText = "This is a Question.";
            subjectChoiceText = "Question: ";
            titleLabel.setText("Question");
            descriptionLabel.setText("Question:");
            part4note.setText("You are totally free in this choice.\nOften, it can be very helpful for others to see your question/answer.");
            currentPart = 2;
        }
        if (e.getSource() == part1Feature) {
            part1Choice = Part1Choice.FEATURE;
            fullChoiceText = "This is a jFed Feature Request.";
            subjectChoiceText = "jFed Feature Request: ";
            part4AllowPublic.setSelected(false);
            titleLabel.setText("jFed Feature Request");
            descriptionLabel.setText("jFed Feature Request:");
            part4note.setText("You are totally free in this choice.\njFed feature requests can sometimes be relevant to share with the list.");
            currentPart = 4;
        }
        if (e.getSource() == part1Bug) {
            part1Choice = Part1Choice.BUG;
            fullChoiceText = "This is a Problem and/or Bug Report.";
            subjectChoiceText = "Bug/Problem Report: ";
            part4AllowPublic.setSelected(false);
            titleLabel.setText("Problem/Bug Report");
            descriptionLabel.setText("Bug description:\n(Please include: expected outcome, steps to reproduce)");
            part4note.setText("You are totally free in this choice.\njFed bug reports and/or testbed problems can sometimes be relevant to share with the list.");
            currentPart = 2;
        }
        if (e.getSource() == part1Other) {
            part1Choice = Part1Choice.OTHER;
            subjectChoiceText = "Feedback: ";
            part4AllowPublic.setSelected(true);
            titleLabel.setText("Feedback");
            descriptionLabel.setText("Feedback:");
            part4note.setText("You are totally free in this choice.");
            currentPart = 2;
        }

        if (part1Choice == null) {
            part1Choice = Part1Choice.OTHER;
            LOG.error("onPartOne() -> Unknown event source! falling back to part1Choice = Part1Choice.OTHER. e.getSource()={}", e.getSource());
        }

        showPart();
    }

    public void onPartTwo(ActionEvent e) {
        if (e.getSource() == part2jFed) {
            part2Choice = Part2Choice.JFED;
            fullChoiceText += "\nThis is about the jFed Experimenter GUI.";
            if (part1Choice == Part1Choice.BUG) {
                fullChoiceText = "This is a jFed Bug Report.";
                subjectChoiceText = "jFed Bug Report: ";
                titleLabel.setText("jFed Bug Report");
                descriptionLabel.setText("Bug description:\n(Please include: expected outcome, steps to reproduce)");
            }
            currentPart = 4;
        }
        if (e.getSource() == part2Testbed) {
            part2Choice = Part2Choice.TESTBED;
            fullChoiceText += "\nThis is about a Testbed.";
            if (part1Choice == Part1Choice.BUG) {
                fullChoiceText = "This is a Testbed Problem Report.";
                subjectChoiceText = "Testbed Problem Report: ";
                titleLabel.setText("Testbed Problem Report");
                descriptionLabel.setText("Problem description:\n(Please include: expected outcome, steps to reproduce)");
            }
            currentPart = 3;
        }
        if (e.getSource() == part2Other) {
            part2Choice = Part2Choice.OTHER;
            currentPart = 3;
        }

        if (part2Choice == null) {
            part2Choice = Part2Choice.OTHER;
            LOG.error("onPartTwo() -> Unknown event source! falling back to part1Choice = Part2Choice.OTHER. e.getSource()={}", e.getSource());
        }

        showPart();
    }

    public void onPartThree(ActionEvent e) {
        if (part3Combobox.getSelectionModel().isEmpty()) {
            part3Choice = null;
        } else {
            part3Choice = part3Combobox.getSelectionModel().getSelectedItem();
            fullChoiceText += "\nThis is about the "+part3Choice.getName()+" Testbed";
        }
        currentPart = 4;

        showPart();
    }

    public void onPartFour(ActionEvent e) {
        part4Choice = part4AllowPublic.isSelected();
        if (part4Choice) {
            fullChoiceText += "\nThis information was shared on the public fed4fire experimenter mailinglist.";
        } else {
            fullChoiceText += "\nThis information was not shared on the public fed4fire experimenter mailinglist.";
        }

        subjectTextField.setText(subjectChoiceText+"YOUR SUBJECT HERE");
        headerTextArea.setText("This message was sent from the jFed Experimenter GUI.\n\n"+fullChoiceText);

        currentPart = 5;
        showPart();
    }

    @FXML
    private void onSaveToFileAction() {
        if (!checkValidity())
            return;

        FileChooser fc = new FileChooser();
        fc.setTitle("Save Bug Report");
        fc.getExtensionFilters().add(new FileChooser.ExtensionFilter("Bug Report (*." + BUGREPORT_EXTENSION + ")", "*" + BUGREPORT_EXTENSION));
        fc.setInitialFileName(String.format("BugReport-%1$tF-%1$tH-%1$tM.xml", new Date()));


        //set the default save dir
        if (System.getProperty("user.home") != null)
            fc.setInitialDirectory(new File(System.getProperty("user.home")));

        if (jFedPreferences.containsPreference(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST)) {
            File dir = new File(jFedPreferences.getString(GuiPreferenceKey.PREF_RSPECFOLDER_REQUEST));
            if (dir.exists()) {
                fc.setInitialDirectory(dir);
            }
        }

        File file = fc.showSaveDialog(submitButton.getScene().getWindow());

        if (file != null) {
            BugReportToFileTask bugReportToFileTask = new BugReportToFileTask(bugReportComposer, file);
            startAndShowBugReportTask(bugReportToFileTask);
        }

    }


    @FXML
    private void onSubmitAction() {
        if (checkValidity()) {
            BugReportToServerTask bugReportToServerTask = new BugReportToServerTask(bugReportComposer);
            startAndShowBugReportTask(bugReportToServerTask);
        }
    }

    private boolean checkValidity() {
        if (descriptionTextArea == null || descriptionTextArea.getText() == null || descriptionTextArea.getText().isEmpty()) {
            JFDialogs.create()
                    .owner(submitButton.getScene().getWindow())
                    .message("Please fill in a description to submit a bug report.")
                    .masthead("Description is a mandatory field")
                    .showError();
            return false;
        }

        if (emailTextField.getText() == null || emailTextField.getText().isEmpty()) {
            JFDialogs.create()
                    .owner(submitButton.getScene().getWindow())
                    .message("You must provide an email address on which we can contact you")
                    .masthead("Email is a mandatory field")
                    .showError();
            return false;
        }      else if(!isValidEmailAddress(emailTextField.getText())){
            JFDialogs.create()
                    .owner(submitButton.getScene().getWindow())
                    .message("Your email address is invalid. Please check your input!")
                    .masthead("Email is a mandatory field")
                    .showError();
            return false;
        }

        return true;
    }

    private static boolean isValidEmailAddress(String email) {
        String ePattern = "^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@((\\[[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\])|(([a-zA-Z\\-0-9]+\\.)+[a-zA-Z]{2,}))$";
        java.util.regex.Pattern p = java.util.regex.Pattern.compile(ePattern);
        java.util.regex.Matcher m = p.matcher(email);
        return m.matches();
    }

    private void startAndShowBugReportTask(final BugReportToFileTask bugReportTask) {
        this.reportTask = bugReportTask;
        submitting.set(true);

        progressLabel.textProperty().bind(bugReportTask.messageProperty());
        submitProgressIndicator.progressProperty().bind(bugReportTask.progressProperty());


        bugReportTask.setOnSucceeded(workerStateEvent -> {
            JFDialogs.create()
                    .owner(submitButton.getScene().getWindow())
                    .message(bugReportTask.getValue())
                    .masthead("Bug report result")
                    .title(null)
                    .showInformation();

            BugReportDialog.this.reportTask = null;
            close();
        });

        bugReportTask.setOnFailed(workerStateEvent -> {
            JFDialogs.create()
                    .owner(submitButton.getScene().getWindow())
                    .message("An error occurred while sending the bug report. Are you connected to the internet?")
                    .masthead("Error while sending bug report")
                    .title("Error while sending bug report")
                    .showException(bugReportTask.getException());

            submitting.set(false);
            progressLabel.textProperty().unbind();
            progressLabel.setText("Processing bug report failed");
            submitProgressIndicator.progressProperty().unbind();
            BugReportDialog.this.reportTask = null;
        });

        new Thread(bugReportTask).start();
    }

    private void close() {
        Stage stage = (Stage) cancelButton.getScene().getWindow();
        if (stage != null)
            stage.close();
    }

    @FXML
    private void onCancelAction() {
        if (reportTask != null) {
            reportTask.cancel();
        }
        close();
    }

    @Override
    public void registerScene(Scene scene) {
        scene.getAccelerators().put(new KeyCodeCombination(KeyCode.F10), () -> saveToFileButton.setVisible(!saveToFileButton.isVisible()));
    }

    @FXML
    private void fetchScreenshot() {
        if (systemClipboard.hasImage()) {
            screenshotImageView.setImage(systemClipboard.getImage());

        } else {
            String osSpecificHelpmessage = "";

            OSDetector.OS currentOS = OSDetector.findOS();
            if (currentOS == OSDetector.OS.WIN) {
                osSpecificHelpmessage =
                        "\t- Press 'Print Screen'/ 'Prnt Scrn' / 'PrtScn',\n" +
                                "\t- Press ALT+'Print Screen' if you only want to make a picture of the active window.";
            } else if (currentOS == OSDetector.OS.MAC) {
                osSpecificHelpmessage =
                        "\t- Press Command-Control-Shift-3 for the whole screen,\n" +
                                "\t- Press Command-Control-Shift-4 if you want to select a specific area of your screen first,\n" +
                                "\t- Press Command-Control-Shift-4 followed by Space if you want to select a specific window.";
            } else {
                osSpecificHelpmessage = "\t- On Ubuntu: press CTRL+'PrtScn'\n" +
                        "\t- On KDE: press 'PrtScn'\n" +
                        "\t- On Gnome: execute 'gnome-screenshot -a -c'";
            }


            JFDialogs.create()
                    .message("The clipboard doesn't contain an image at the moment. " +
                            "To include a screenshot in the bugreport, make a screenshot of the relevant windows, then press the button in the Bug Report dialog.\n\nTo make a screenshot and store it on the clipboard: \n" + osSpecificHelpmessage)
                    .masthead("Fetching screenshot from clipboard failed")
                    .showError();
        }


    }
}
